#include <omp.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>

#define V1
//#define V2

#ifdef V1
int num1, num2;
#endif

#ifdef V2
double *A;
double *B;
double *C;
#endif

#define PRINT1
#define PRINT2

void work(int i, int j)
{
  printf("\ni: %d   j: %d   thread: %d", i, j, omp_get_thread_num());
}

void good_nesting(int n)
{
  int i, j;

  omp_set_num_threads(2);
  omp_set_nested(true);

  bool f = omp_get_nested();
  if (f)
    printf("\nnested");
  else
    printf("\nnot nested");
  #pragma omp parallel private(j)
  {
#ifdef PRINT1
    #pragma omp master
      printf("\nnum of threads = %d", omp_get_num_threads());
    #pragma omp critical
    {
      num1 = omp_get_thread_num();
      printf("\nthread is #%d", num1);
    }
#endif
    #pragma omp for
    for (i = 0; i < n; i++)
    {
      #pragma omp parallel
      {
#ifdef PRINT2
        #pragma omp master
          printf("\nnum of threads = %d", omp_get_num_threads());
        #pragma omp critical
        {
          num2 = omp_get_thread_num();
          printf("\nthread #%d", num2);
        }
#endif

        #pragma omp for
        for (j = 0; j < n; j++)
          work(i, j);
      }
    }
  }
}

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
#ifdef V1
  good_nesting(4);
#endif
#ifdef V2
  double tmp;
  int i, j, k;
  long tck;
  clock_t begin, end;
  int M;

  if (argc != 2)
    printf("Usage: matr #size");
  else
  {
    M = atoi(argv[1]);
  //randomize();
    printf("\nMatrix multiplication: OpenMP variant\n");
    printf("  Matrix size = %i\n", M);
    A = (double *)malloc(M * M * sizeof(double));
    B = (double *)malloc(M * M * sizeof(double));
    C = (double *)malloc(M * M * sizeof(double));
    for (i = 0; i < M; i++)
      for (j = 0; j < M; j++)
      {
        A[i * M + j] = 1;//((double)rand() / (double)RAND_MAX);
        B[i * M + j] = 1;//((double)rand() / (double)RAND_MAX);
        C[i * M + j] = 0.0;
      }
    begin = clock();
/*
#pragma omp parallel
    {
      #pragma omp for private(i, j, k)
      for (i = 0; i < M; i++)
        #pragma omp for private(j, k)
        for (j = 0; j < M; j++)
          for (k = 0; k < M; k++)
            C[i * M + j] += A[i * M + k] * B[k * M + j];
    }
*/
    omp_set_nested(true);
    #pragma omp parallel for private(i, j, k)
    for (i = 0; i < M; i++)
      #pragma omp parallel for private(j, k)
      for(j = 0; j < M; j++)
        for(k = 0; k < M; k++)
          C[i * M + j] += A[i * M + k] * B[k * M + j];
    end = clock();
    tck = end - begin;
    printf("\n");
    for (i = 0; i < M; i++)
    {
      for(j = 0; j < M; j++)
        printf("c[%d][%d] = %.0f  ", i, j, C[i * M + j]);
      printf("\n");
    }

    printf("  time:\n");
    printf("    seconds = %f",((float)tck)/CLOCKS_PER_SEC);
    free(C);
    free(B);
    free(A);  
  }
#endif
  printf("\n");
  system("pause");

}
//---------------------------------------------------------------------------
